"""
The fixtures module provides classes returned by fixtures registered by pytest in snappylapy.

Snappylapy provides the following fixtures.

- expect: Expect
    - Allows for validating various expectations on the test results and do snapshot testing.
- load_snapshot: LoadSnapshot
    - Allows loading from a snapshot created by another test.
"""

from __future__ import annotations

from .expectation_classes import (
    BytesExpect,
    DataframeExpect,
    DictExpect,
    ListExpect,
    ObjectExpect,
    StringExpect,
)
from .models import Settings
from .serialization import (
    BytesSerializer,
    JsonPickleSerializer,
    StringSerializer,
)
from snappylapy.constants import DIRECTORY_NAMES
from snappylapy.session import SnapshotSession
from typing import Any, Protocol, overload


class _CallableExpectation(Protocol):
    """Protocol for callable expectations to use internally in this module."""

    def __call__(
        self,
        data_to_snapshot: Any,  # noqa: ANN401
        name: str | None = None,
        filetype: str = "snapshot.txt",
    ) -> DictExpect | ListExpect | StringExpect | BytesExpect | DataframeExpect:
        """Call the expectation with the given parameters."""
        ...


class Expect:
    """
    Snapshot testing fixture class.

    Do not instantiate this class directly, instead use the `expect` fixture provided by pytest.
    Use this class as a type hint for the `expect` fixture.
    """

    def __init__(
        self,
        snappylapy_session: SnapshotSession,
        snappylapy_settings: Settings,
    ) -> None:
        """"""  # noqa: D419, blank, since used in doc generation
        self.settings = snappylapy_settings

        self.dict = DictExpect(self.settings, snappylapy_session)
        """DictExpect instance for configuring snapshot testing of dictionaries.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`dict`): The dictionary data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot".
        - `filetype` (`str`, optional): The file type of the snapshot, by default "dict.json".

        Returns
        -------
        DictExpect
        - `DictExpect`: The instance of the DictExpect class.

        Example
        -------
        `test_fixture_expect_dict.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_dict(expect: Expect) -> None:
            data: dict[str, str] = {"key": "value"}
            expect.dict(data).to_match_snapshot()
            expect.dict(data, name="snapshot_name", filetype="dict.json").to_match_snapshot()
        ```
        """

        self.list = ListExpect(self.settings, snappylapy_session)
        """ListExpect instance for configuring snapshot testing of lists.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`list`): The list data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "list.json".

        Returns
        -------
        ListExpect
        - `ListExpect`: The instance of the ListExpect class.

        Example
        -------
        `test_fixture_expect_list.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_list(expect: Expect) -> None:
            data: list[int] = [1, 2, 3]
            expect.list(data).to_match_snapshot()
        ```
        """

        self.string = StringExpect(self.settings, snappylapy_session)
        """StringExpect instance for configuring snapshot testing of strings.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`str`): The string data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "string.txt".

        Returns
        -------
        StringExpect
        - `StringExpect`: The instance of the StringExpect class.

        Example
        -------
        `test_fixture_expect_string.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_string(expect: Expect) -> None:
            data: str = "Hello, World!"
            expect.string(data).to_match_snapshot()
        ```
        """

        self.bytes = BytesExpect(self.settings, snappylapy_session)
        """BytesExpect instance for configuring snapshot testing of bytes.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`bytes`): The bytes data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "bytes.txt".

        Returns
        -------
        BytesExpect
        - `BytesExpect`: The instance of the BytesExpect class.

        Example
        -------
        `test_fixture_expect_bytes.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_bytes(expect: Expect) -> None:
            data: bytes = b"binary data"
            expect.bytes(data).to_match_snapshot()
        ```
        """

        self.dataframe = DataframeExpect(self.settings, snappylapy_session)
        """DataframeExpect instance for configuring snapshot testing of dataframes.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`pd.DataFrame`): The dataframe data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "dataframe.json".

        Returns
        -------
        DataframeExpect
        - `DataframeExpect`: The instance of the DataframeExpect class.

        Example
        -------
        `test_fixture_expect_dataframe.py`
        ```python
        import pytest
        import pandas as pd
        from snappylapy.fixtures import Expect

        def test_expect_dataframe(expect: Expect) -> None:
            df: pd.DataFrame = pd.DataFrame({"key": ["value1", "value2"]})
            expect.dataframe(df).to_match_snapshot()
        ```
        """

        self.object = ObjectExpect(self.settings, snappylapy_session)
        """ObjectExpect instance for configuring snapshot testing of generic objects.

        The instance is callable with the following parameters:

        Parameters
        ----------
        - `data_to_snapshot` (`object`): The object data to be snapshotted.
        - `name` (`str`, optional): The name of the snapshot.
        - `filetype` (`str`, optional): The file type of the snapshot, by default "object.json".

        Returns
        -------
        ObjectExpect
        - `ObjectExpect`: The instance of the ObjectExpect class.

        Example
        -------
        `test_fixture_expect_object.py`
        ```python
        import pytest
        from snappylapy.fixtures import Expect

        def test_expect_object(expect: Expect) -> None:
            obj: dict[str, str] = {"key": "value"}
            expect.object(obj).to_match_snapshot()
        ```
        """

    @overload
    def __call__(self, data_to_snapshot: dict, name: str | None = None, filetype: str | None = None) -> DictExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: list[Any],
        name: str | None = None,
        filetype: str | None = None,
    ) -> ListExpect: ...

    @overload
    def __call__(self, data_to_snapshot: str, name: str | None = None, filetype: str | None = None) -> StringExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: bytes,
        name: str | None = None,
        filetype: str | None = None,
    ) -> BytesExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: DataframeExpect.DataFrame,
        name: str | None = None,
        filetype: str | None = None,
    ) -> DataframeExpect: ...

    @overload
    def __call__(
        self,
        data_to_snapshot: Any,  # noqa: ANN401
        name: str | None = None,
        filetype: str | None = None,
    ) -> ObjectExpect: ...

    def __call__(
        self,
        data_to_snapshot: dict | list[Any] | str | bytes | DataframeExpect.DataFrame,
        name: str | None = None,
        filetype: str | None = None,
    ) -> DictExpect | ListExpect | StringExpect | BytesExpect | DataframeExpect | ObjectExpect:
        """Call the fixture with the given parameters. Falls back to object handler for custom objects."""
        kwargs: dict[str, str] = {}
        if name is not None:
            kwargs["name"] = name
        if filetype is not None:
            kwargs["filetype"] = filetype

        type_map: dict[type, _CallableExpectation] = {
            dict: self.dict,
            list: self.list,
            str: self.string,
            bytes: self.bytes,
            DataframeExpect.DataFrame: self.dataframe,
        }

        for typ, func in type_map.items():
            if isinstance(typ, type) and isinstance(data_to_snapshot, typ):
                return func(data_to_snapshot, **kwargs)

        # Check if the object is a pandas DataFrame without importing pandas directly
        if (
            type(data_to_snapshot).__module__.startswith("pandas")
            and type(data_to_snapshot).__name__ == "DataFrame"
            # TODO: Create a protocol class instead that contains all the dependencies we are depending on
        ):
            return self.dataframe(data_to_snapshot, **kwargs)  # type: ignore[arg-type]

        # Fallback: treat custom objects as dicts for snapshotting
        return self.object(data_to_snapshot, **kwargs)

    def _read_snapshot(self) -> bytes:
        """Read the snapshot file."""
        return (self.settings.snapshot_dir / self.settings.filename).read_bytes()

    def _read_test_results(self) -> bytes:
        """Read the test results file."""
        return (self.settings.test_results_dir / self.settings.filename).read_bytes()


class LoadSnapshot:
    """
    Snapshot loading class.

    This class provides methods to load snapshots created by other tests.
    Each method loads and deserializes a specific type of snapshot.
    """

    def __init__(self, settings: Settings) -> None:
        """Do not initialize the LoadSnapshot class directly, should be used through the `load_snapshot` fixture in pytest."""  # noqa: E501
        self.settings = settings

    def _read_snapshot(self) -> bytes:
        """Read the snapshot file."""
        if not self.settings.depending_snapshots_base_dir:
            msg = "Depending snapshots base directory is not set."
            raise ValueError(msg)
        return (
            self.settings.depending_snapshots_base_dir
            / DIRECTORY_NAMES.snapshot_dir_name
            / self.settings.depending_filename
        ).read_bytes()

    def dict(self) -> dict[str, int]:
        """
        Load dictionary snapshot.

        Use this method to load a dictionary snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_dict.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def create_dict() -> dict[str, int]:
            return {"apples": 3, "bananas": 5}

        def test_save_dict_snapshot(expect: Expect) -> None:
            data: dict[str, int] = create_dict()
            expect(data).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_dict_snapshot])
        def test_load_snapshot_dict(load_snapshot: LoadSnapshot) -> None:
            data: dict[str, int] = load_snapshot.dict()
            assert data["apples"] == 3
            assert data["bananas"] == 5
        ```
        """
        self.settings.depending_filename_extension = "dict.json"
        return JsonPickleSerializer[dict]().deserialize(self._read_snapshot())

    def list(self) -> list[Any]:
        """
        Load list snapshot.

        Use this method to load a list snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_list.py`
        ```python
        import pytest
        from typing import Any
        from snappylapy import LoadSnapshot, Expect

        def transform_data(data: list) -> list:
            return [x * 2 for x in data]

        def next_transformation(data: list) -> list:
            return [x + 1 for x in data]

        def test_transform_data(expect: Expect) -> None:
            data = [1, 2, 3]
            result = transform_data(data)
            expect(result).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_transform_data])
        def test_next_transformation(load_snapshot: LoadSnapshot, expect: Expect) -> None:
            data: list[Any] = load_snapshot.list()
            result = next_transformation(data)
            expect(result).to_match_snapshot()
        ```
        """
        self.settings.depending_filename_extension = "list.json"
        return JsonPickleSerializer[list[Any]]().deserialize(self._read_snapshot())

    def string(self) -> str:
        """
        Load string snapshot.

        Use this method to load a string snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_string.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_string_snapshot(expect: Expect) -> None:
            message: str = "Hello, pytest!"
            expect(message).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_string_snapshot])
        def test_load_snapshot_string(load_snapshot: LoadSnapshot) -> None:
            data: str = load_snapshot.string()
            assert data == "Hello, pytest!"
        ```
        """
        self.settings.depending_filename_extension = "string.txt"
        return StringSerializer().deserialize(self._read_snapshot())

    def bytes(self) -> bytes:
        r"""
        Load bytes snapshot.

        Use this method to load a bytes snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_bytes.py`
        ```python
        import pytest
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_bytes_snapshot(expect: Expect) -> None:
            data: bytes = b"\x01\x02\x03"
            expect(data).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_bytes_snapshot])
        def test_load_snapshot_bytes(load_snapshot: LoadSnapshot) -> None:
            data: bytes = load_snapshot.bytes()
            assert data == b"\x01\x02\x03"
        ```
        """
        self.settings.depending_filename_extension = "bytes.txt"
        return BytesSerializer().deserialize(self._read_snapshot())

    def dataframe(self) -> DataframeExpect.DataFrame:
        """
        Load dataframe snapshot.

        Use this method to load a dataframe snapshot that was created in a previous test.
        This is useful for reusing test data, isolating dependencies, and verifying integration between components.

        Example usage:
        --------------
        `test_load_snapshot_from_file_dataframe.py`
        ```python
        import pytest
        import pandas as pd
        from snappylapy.fixtures import LoadSnapshot, Expect

        def test_save_dataframe_snapshot(expect: Expect) -> None:
            df: pd.DataFrame = pd.DataFrame({"numbers": [1, 2, 3]})
            expect(df).to_match_snapshot()

        @pytest.mark.snappylapy(depends=[test_save_dataframe_snapshot])
        def test_load_snapshot_dataframe(load_snapshot: LoadSnapshot) -> None:
            df: pd.DataFrame = load_snapshot.dataframe()
            assert df["numbers"].sum() == 6
        ```
        """
        self.settings.depending_filename_extension = "dataframe.json"
        return DataframeExpect.DataFrame(
            JsonPickleSerializer[DataframeExpect.DataFrame]().deserialize(self._read_snapshot()),
        )



import pathlib
from snappylapy import Expect, LoadSnapshot
import json
import pytest

def test_snapshot_string(expect: Expect):
    """Test snapshot with string data."""
    expect.string("Hello World").to_match_snapshot()

def test_custom_type(expect: Expect):
    """Test snapshot with a custom type."""
    class CustomObject:
        """A custom type for snapshot."""
        pass

    expect(CustomObject()).to_match_snapshot()

def test_snapshot_bytes(expect: Expect):
    """Test snapshot with bytes data."""
    expect.bytes(b"Hello World", name="bytes_snapshot").to_match_snapshot()

def test_snapshot_dict(expect: Expect):
    """Test snapshot with dictionary data."""
    expect.dict({
        "name": "John Doe",
        "age": 31
    }).to_match_snapshot()

def test_snapshot_list(expect: Expect):
    """Test snapshot with list data."""
    expect.list(["John Doe", 31]).to_match_snapshot()

def test_snapshot_json_bytes(expect: Expect):
    """Test snapshot with JSON bytes data."""
    data = json.dumps({"name": "John Doe", "age": 31}).encode()
    expect.bytes(data, name="json_bytes_snapshot").to_match_snapshot()

@pytest.mark.snappylapy(depends=[test_snapshot_dict])
def test_load_snapshot_from_file(load_snapshot: LoadSnapshot):
    """Test loading snapshot data created in test_snapshot_dict from a file using the deserializer."""
    data = load_snapshot.dict()
    assert data == {"name": "John Doe", "age": 31}

@pytest.mark.snappylapy(depends=[test_snapshot_string])
def test_load_snapshot_from_file_string(load_snapshot: LoadSnapshot):
    """Test loading snapshot data created in test_snapshot_string from a file using the deserializer."""
    data = load_snapshot.string()
    assert data == "Hello World"


def test_snapshot_python_code(expect: Expect):
    """Test snapshot with Python code string."""
    py_code = "print('Hello World')"
    expect.string(py_code, filetype="py", name="python_code_snapshot").to_match_snapshot()

@pytest.mark.snappylapy(output_dir="custom_dir")
def test_snapshot_with_custom_directories(expect: Expect):
    """Test snapshot with custom directories."""
    expect.string("Hello World").to_match_snapshot()

@pytest.mark.snappylapy(depends=[test_snapshot_with_custom_directories])
def test_load_snapshot_from_custom_dir(load_snapshot: LoadSnapshot):
    """Test loading snapshot data created in test_snapshot_with_custom_directories from a file using the deserializer."""
    data = load_snapshot.string()
    assert data == "Hello World"

def test_snapshot_multiple_assertions(expect: Expect):
    """Test snapshot with multiple assertions."""
    expect.string("Hello World").to_match_snapshot()
    expect.dict({
        "name": "John Doe",
        "age": 31
    }).to_match_snapshot()

@pytest.mark.parametrize("data", [
    "Hello World",
    "Hello Galaxy",
    "Hello Universe",
])
def test_snapshot_parametrized(data: str, expect: Expect):
    """Test snapshot with parametrized data."""
    expect.string(data).to_match_snapshot()


@pytest.mark.snappylapy(foreach_folder_in="test_data")
def test_snapshot_multiple_folders_snappylapy_marker(test_directory: pathlib.Path, expect: Expect):
    """Test snapshot with multiple folders."""
    expect.string("Hello World").to_match_snapshot()

@pytest.mark.parametrize("test_directory", list(pathlib.Path("test_data").iterdir()), ids=lambda x: x.name)
def test_snapshot_multiple_folders_pytest_parametrize(test_directory: pathlib.Path, expect: Expect):
    """Test snapshot with multiple folders."""
    expect.string("Hello World").to_match_snapshot()

@pytest.mark.snappylapy(depends=[test_snapshot_multiple_folders_pytest_parametrize])
@pytest.mark.skip(reason="Functionaility not implemented yet.")
def test_load_parametrized_snapshot_from_file(load_snapshot: LoadSnapshot):
    """Test loading snapshot data created in test_snapshot_parametrized from a file using the deserializer."""
    data = load_snapshot.string()
    assert data == "Hello World"

@pytest.mark.snappylapy(depends=[test_snapshot_multiple_folders_snappylapy_marker])
def test_load_snapshot_from_multiple_folders(load_snapshot: LoadSnapshot):
    """Test loading snapshot data created in test_snapshot_multiple_folders from a file using the deserializer."""
    data = load_snapshot.string()
    assert data == "Hello World"


"""Utility functions for handling directories in snappylapy."""
import pathlib
from snappylapy.constants import DIRECTORY_NAMES


def find_directories(directory_names: list[str]) -> list[pathlib.Path]:
    """Find directories with the given names."""
    found_dirs: list[pathlib.Path] = []
    for dir_name in directory_names:
        found_dirs.extend(pathlib.Path().rglob(dir_name))
    return found_dirs


def get_file_paths_from_directories(list_of_directories: list[pathlib.Path]) -> list[pathlib.Path]:
    """Get file paths from directories."""
    list_of_files_to_delete: list[pathlib.Path] = []
    for directory in list_of_directories:
        if not directory.is_dir():
            error_msg = f"{directory} is not a directory."
            raise ValueError(error_msg)
        list_of_files_to_delete.extend(file for file in directory.iterdir() if file.is_file())
    return list_of_files_to_delete


class DirectoryNamesUtil:
    """
    Utility class to handle directory names and operations related to them.

    This class extends the DirectoryNames class to provide methods for finding directories
    and file paths created by snappylapy.
    """

    def get_all_directory_names(self) -> list[str]:
        """Get all directory names."""
        return [DIRECTORY_NAMES.snapshot_dir_name, DIRECTORY_NAMES.test_results_dir_name]

    def get_all_directories_for_test_results(self) -> list[pathlib.Path]:
        """Get all directories for test results."""
        return find_directories([DIRECTORY_NAMES.test_results_dir_name])

    def get_all_directories_for_snapshots(self) -> list[pathlib.Path]:
        """Get all directories for snapshots."""
        return find_directories([DIRECTORY_NAMES.snapshot_dir_name])

    def get_all_file_paths_test_results(self) -> list[pathlib.Path]:
        """Get all file paths in the test results directory."""
        directories = find_directories([DIRECTORY_NAMES.test_results_dir_name])
        return get_file_paths_from_directories(directories)

    def get_all_file_paths_snapshots(self) -> list[pathlib.Path]:
        """Get all file paths in the snapshot directory."""
        directories = find_directories([DIRECTORY_NAMES.snapshot_dir_name])
        return get_file_paths_from_directories(directories)

    def get_all_directories_created_by_snappylapy(self) -> list[pathlib.Path]:
        """Get all directories created by snappylapy."""
        return find_directories(self.get_all_directory_names())

    def get_all_file_paths_created_by_snappylapy(self) -> list[pathlib.Path]:
        """Get all file paths created by snappylapy."""
        return get_file_paths_from_directories(
            self.get_all_directories_created_by_snappylapy(),
        )



"""Snapshot testing and expectations for strings."""
from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import StringSerializer


class StringExpect(BaseSnapshot[str]):
    """Snapshot testing for strings."""

    serializer_class = StringSerializer

    def __call__(self,
                 data_to_snapshot: str,
                 name: str | None = None,
                 filetype: str = "string.txt") -> StringExpect:
        """Prepare a string for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""Testing of the string test expectations class."""
import pytest
import pathlib
import snappylapy.expectation_classes.expect_dataframe as module_on_test
from snappylapy.models import Settings
from unittest import mock

try:
    import pandas as pd  # noqa: F401
    pandas_installed = True
except ImportError:
    pandas_installed = False

@pytest.fixture
def expect() -> module_on_test.DataframeExpect:
    snappylapy_session = mock.MagicMock()
    return module_on_test.DataframeExpect(
        settings=Settings(
            test_filename="test_file",
            test_function="test_function",
        ),
        snappylapy_session=snappylapy_session,
    )

# @pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
# def test_to_match_snapshot(expect: module_on_test.DataframeExpect) -> None:
#     """Test the to_match_snapshot method of DataframeExpect."""
#     # TODO This will not actually create a snapshot with the snapshot session
#     dataframe_to_test = pd.DataFrame({"key": ["value1", "value2"]})
#     expect(dataframe_to_test).to_match_snapshot()

@pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
def test_columns_not_to_contain_nulls(expect: module_on_test.DataframeExpect) -> None:
    """Test the columns_not_to_contain_nulls method of DataframeExpect."""
    dataframe_to_test = pd.DataFrame({"key": ["value1", "value2"]})
    expect(dataframe_to_test).columns_not_to_contain_nulls()

@pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
def test_column_not_to_contain_nulls_fails(expect: module_on_test.DataframeExpect) -> None:
    """Test the column_not_to_contain_nulls method of DataframeExpect with failure."""
    dataframe_to_test = pd.DataFrame({"key": ["value1", "value2", None]})
    with pytest.raises(ValueError, match="Column key contains 1 null values"):
        expect(dataframe_to_test).column_not_to_contain_nulls("key")

@pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
def test_multiple_columns_not_to_contain_nulls_fails(expect: module_on_test.DataframeExpect) -> None:
    """Test the columns_not_to_contain_nulls method of DataframeExpect with failure."""
    dataframe_to_test = pd.DataFrame({"key1": ["value1", "value2"], "key2": [None, "value4"]})
    with pytest.raises(ValueError, match="Column key2 contains 1 null values"):
        expect(dataframe_to_test).columns_not_to_contain_nulls(["key1", "key2"])

@pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
def test_column_not_to_contain_nulls_fails_no_data(expect: module_on_test.DataframeExpect) -> None:
    """Test the column_not_to_contain_nulls method of DataframeExpect with no data."""
    with pytest.raises(ValueError, match="No data to check. Call __call__ first."):
        expect.column_not_to_contain_nulls("key")

@pytest.mark.skipif(not pandas_installed, reason="pandas is not installed")
def test_columns_to_match_regex(
    expect: module_on_test.DataframeExpect,
) -> None:
    """Test the columns_to_match_regex method of DataframeExpect."""
    dataframe_to_test = pd.DataFrame({"key1": ["value1", "value2"], "key2": ["test1", "test2"]})
    expect(dataframe_to_test).columns_to_match_regex({"key1": r"value\d", "key2": r"test\d"})



"""Read-only constants."""

import pathlib
from typing import NamedTuple

DEFAULT_SNAPSHOT_BASE_DIR = pathlib.Path()
OUTPUT_JSON_INDENTATION_LEVEL = 2


class DirectoryNames(NamedTuple):
    """Immutable directory names for snappylapy."""

    snapshot_dir_name: str
    test_results_dir_name: str


DIRECTORY_NAMES: DirectoryNames = DirectoryNames(
    snapshot_dir_name="__snapshots__", test_results_dir_name="__test_results__",
)



"""Pytest plugin for snapshot testing."""

from __future__ import annotations

import os
import re
import pytest
import pathlib
import _pytest.mark
from collections.abc import Callable
from snappylapy import Expect, LoadSnapshot
from snappylapy._utils_directories import DirectoryNamesUtil
from snappylapy.constants import DEFAULT_SNAPSHOT_BASE_DIR
from snappylapy.exceptions import TestDirectoryNotParametrizedError
from snappylapy.fixtures import Settings
from snappylapy.session import SnapshotSession
from typing import Any


def _get_kwargs_from_depend_function(
    depends_function: Callable,
    marker_name: str,
    kwags_key: str,
) -> str | None:
    """Get a test function with a pytest marker assigned and get a value from the marker."""
    if not hasattr(depends_function, "pytestmark"):
        return None
    marks: list[_pytest.mark.structures.Mark] = depends_function.pytestmark
    for mark in marks:
        if mark.name == marker_name:
            return mark.kwargs.get(kwags_key, None)
    return None


def _get_args_from_depend_function(
    depends_function: Callable,
    marker_name: str,
) -> tuple[Any] | None:
    """Get a test function with a pytest marker assigned and get a value from the marker."""
    if not hasattr(depends_function, "pytestmark"):
        return None
    marks: list[_pytest.mark.structures.Mark] = depends_function.pytestmark
    for mark in marks:
        if mark.name == marker_name:
            return mark.args
    return None


@pytest.fixture
def snappylapy_settings(request: pytest.FixtureRequest) -> Settings:
    """Initialize the Settings object for the test."""
    update_snapshots = request.config.getoption("--snapshot-update")
    marker = request.node.get_closest_marker("snappylapy")
    match = re.search(r"\[(.*?)\]", request.node.name)
    param_name: str | None = match.group(1) if match else None
    settings = Settings(
        test_filename=request.module.__name__,
        test_function=request.node.originalname,
        custom_name=param_name,
        snapshot_update=update_snapshots,
    )
    if marker:
        output_dir: str | pathlib.Path = marker.kwargs.get("output_dir", None)
        if output_dir:
            settings.snapshots_base_dir = pathlib.Path(output_dir)
    path_output_dir: pathlib.Path | None = None
    if hasattr(request, "param"):
        path_output_dir = request.param
        if path_output_dir is None:
            # TODO: Add a better error message
            msg = "Path output directory cannot be None"
            raise ValueError(msg)
        settings.depending_snapshots_base_dir = pathlib.Path(path_output_dir)
        settings.snapshots_base_dir = pathlib.Path(path_output_dir)
        settings.custom_name = path_output_dir.name
    # If not parametrized, get the depends from the marker
    depends: list = marker.kwargs.get("depends", []) if marker else []
    if depends:
        input_dir_from_depends = _get_kwargs_from_depend_function(depends[0], "snappylapy", "output_dir")
        if input_dir_from_depends:
            path_output_dir = pathlib.Path(input_dir_from_depends)
        settings.depending_test_filename = depends[0].__module__
        settings.depending_test_function = depends[0].__name__
    settings.depending_snapshots_base_dir = path_output_dir or DEFAULT_SNAPSHOT_BASE_DIR
    return settings


@pytest.fixture
def expect(request: pytest.FixtureRequest, snappylapy_settings: Settings) -> Expect:
    """Initialize the snapshot object with update_snapshots flag from pytest option."""
    snappylapy_session: SnapshotSession = request.config.snappylapy_session  # type: ignore[attr-defined]
    return Expect(
        snappylapy_session=snappylapy_session,
        snappylapy_settings=snappylapy_settings,
    )


@pytest.fixture
def load_snapshot(snappylapy_settings: Settings) -> LoadSnapshot:
    """Initialize the LoadSnapshot object."""
    return LoadSnapshot(snappylapy_settings)


@pytest.hookimpl(tryfirst=True)
def pytest_collection_modifyitems(
    session: pytest.Session,
    config: pytest.Config,
    items: list[pytest.Function],
) -> None:
    """Sort the tests based on the dependencies."""
    del config, session  # Unused
    for item in items:
        marker = item.get_closest_marker("snappylapy")
        if not marker:
            continue
        depends = marker.kwargs.get("depends", [])
        for depend in depends:
            for i, test in enumerate(items):
                if test.function != depend:
                    continue
                # Check if it is already earlier in the list than the dependency
                if i < items.index(item):
                    # Preserve the original order
                    break
                # Move the test to the position after the dependency
                items.insert(i + 1, items.pop(items.index(item)))
                break


def pytest_configure(config: pytest.Config) -> None:
    """Register the markers used."""
    config.addinivalue_line(
        "markers",
        "snappylapy: mark test to load snapshot data from a file.",
    )


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add the CLI option for updating snapshots."""
    group = parser.getgroup("terminal reporting")
    group.addoption(
        "--snapshot-update",
        action="store_true",
        dest="snapshot_update",
        default=False,
        help="update snapshots.",
    )


def pytest_sessionstart(session: pytest.Session) -> None:
    """Initialize the snapshot session before running tests."""
    # Check if we're in discovery/collection mode
    if getattr(session.config.option, "collectonly", False) or getattr(session.config.option, "collect_only", False):
        return

    session.config.snappylapy_session = SnapshotSession()  # type: ignore[attr-defined]
    directory_util: DirectoryNamesUtil = DirectoryNamesUtil()
    files_to_delete: list[pathlib.Path] = directory_util.get_all_file_paths_test_results()
    for file_path in files_to_delete:
        file_path.unlink()


class ExceptionDuringTestSetupError(Exception):
    """Error raised when an exception is raised during the setup of the tests."""


class ReturnError:
    """Return an error when trying to access the attribute during the setup of the tests."""

    def __init__(self, exception: Exception, message: str) -> None:
        self._message = message
        self._exception = exception

    def __getattribute__(self, name: str) -> Any:  # noqa: ANN401
        """Raise an exception when trying to access the attribute, if exception was raised during the setup of tests."""
        exception = object.__getattribute__(self, "_exception")
        if exception is not None and os.getenv("PYTEST_CURRENT_TEST"):
            exception_message = f"When during setup of the tests an error was raised: {exception}"
            if self._message:
                exception_message = f"{self._message} {exception_message}"
            raise ExceptionDuringTestSetupError from exception
        return object.__getattribute__(self, name)


@pytest.fixture
def test_directory(snappylapy_settings: Settings) -> pathlib.Path:
    """Get the test directory for the test. Raise a better error message if the fixture is not parametrized."""
    try:
        return pathlib.Path(snappylapy_settings.snapshots_base_dir)
    except Exception as e:
        raise TestDirectoryNotParametrizedError from e


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Generate parametrized tests for the pipeline output and input."""
    marker = metafunc.definition.get_closest_marker("snappylapy")
    if not marker:
        return
    foreach_folder_in: str | pathlib.Path | None = marker.kwargs.get("foreach_folder_in", None)
    if foreach_folder_in:
        test_cases = [p for p in pathlib.Path(foreach_folder_in).iterdir() if p.is_dir()]
        ids = [p.name for p in test_cases]
        metafunc.parametrize("snappylapy_settings", test_cases, indirect=True, ids=ids)
    depends = marker.kwargs.get("depends", []) if marker else []
    if depends:
        function_depends = marker.kwargs["depends"][0]
        if not hasattr(function_depends, "pytestmark"):
            return
        function_depends_marker: _pytest.mark.structures.Mark = function_depends.pytestmark[0]
        # It might be parametrized
        # Example: Mark(name='parametrize', args=('test_directory', ['test_data/case1', 'test_data/case2']), kwargs={})
        # Parametize the snappylapy_settings fixture
        # if function_depends_marker.name == "parametrize":
        #     ids = function_depends_marker.kwargs.get("ids", None)
        #     metafunc.parametrize("snappylapy_settings", function_depends_marker.args[1], indirect=True, ids=ids)
        if function_depends_marker.name == "snappylapy":
            foreach_folder_in = _get_kwargs_from_depend_function(depends[0], "snappylapy", "foreach_folder_in")
            if not foreach_folder_in:
                return
            test_cases = [p for p in pathlib.Path(foreach_folder_in).iterdir() if p.is_dir()]
            ids = [p.name for p in test_cases]
            metafunc.parametrize("snappylapy_settings", test_cases, indirect=True, ids=ids)



"""Snapshot testing and expectations for bytes."""

from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import BytesSerializer


class BytesExpect(BaseSnapshot[bytes]):
    """Snapshot testing for bytes."""

    serializer_class = BytesSerializer

    def __call__(self, data_to_snapshot: bytes, name: str | None = None, filetype: str = "bytes.txt") -> BytesExpect:
        """Prepare bytes for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""Snapshot testing and expectations for lists."""
from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import JsonPickleSerializer
from typing import Any


class ListExpect(BaseSnapshot[list[Any]]):
    """Snapshot testing for lists."""

    serializer_class = JsonPickleSerializer[list[Any]]

    def __call__(
        self,
        data_to_snapshot: list[Any],
        name: str | None = None,
        filetype: str = "list.json",
    ) -> ListExpect:
        """Prepare a list for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""Models for snappylapy."""

from __future__ import annotations

import pathlib
from dataclasses import dataclass
from snappylapy.constants import DIRECTORY_NAMES


@dataclass
class Settings:
    """Shared setting for all the strategies for doing snapshot testing."""

    test_filename: str
    """Filename of the test module where the current test are defined."""

    test_function: str
    """Name of the test function."""

    custom_name: str | None = None
    """Custom name for the snapshot file."""

    snapshots_base_dir: pathlib.Path = pathlib.Path()
    """Output base directory for storing snapshots."""

    snapshot_update: bool = False
    """Flag to update the snapshots."""

    filename_extension: str = "txt"
    """Extension for the output of snapshot file."""

    # Configurations for depending
    depending_test_filename: str | None = None
    """Filename of the test module where the depending test are defined. Used for loading."""

    depending_test_function: str | None = None
    """Name of the depending test function. Used for loading."""

    depending_filename_extension: str | None = None
    """Extension of the depending snapshot file. Used for loading."""

    depending_snapshots_base_dir: pathlib.Path | None = None
    """Input base directory for loading snapshots."""

    @property
    def snapshot_dir(self) -> pathlib.Path:
        """Get the snapshot directory."""
        return pathlib.Path(self.snapshots_base_dir) / DIRECTORY_NAMES.snapshot_dir_name

    @property
    def test_results_dir(self) -> pathlib.Path:
        """Get the test results directory."""
        return pathlib.Path(self.snapshots_base_dir) / DIRECTORY_NAMES.test_results_dir_name

    @property
    def filename(self) -> str:
        """Get the snapshot filename."""
        if self.custom_name is not None:
            return f"[{self.test_filename}][{self.test_function}][{self.custom_name}].{self.filename_extension}"
        return f"[{self.test_filename}][{self.test_function}].{self.filename_extension}"

    @property
    def depending_filename(self) -> str:
        """Get the depending snapshot filename."""
        if (
            not self.depending_test_filename
            or not self.depending_test_function
            or not self.depending_filename_extension
        ):
            msg = "Missing depending test filename, function or extension."
            raise ValueError(msg)
        if self.custom_name is not None:
            return f"[{self.depending_test_filename}][{self.depending_test_function}][{self.custom_name}].{self.depending_filename_extension}"  # noqa: E501
        return f"[{self.depending_test_filename}][{self.depending_test_function}].{self.depending_filename_extension}"



"""Snapshot testing and expectations for dicts."""
from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import JsonPickleSerializer


class DictExpect(BaseSnapshot[dict]):
    """Snapshot testing for dictionaries."""

    serializer_class = JsonPickleSerializer[dict]

    def __call__(self,
                 data_to_snapshot: dict,
                 name: str | None = None,
                 filetype: str = "dict.json") -> DictExpect:
        """Prepare a dictionary for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""Test cases for serialization module."""
from snappylapy.serialization import StringSerializer, JsonSerializer, JsonPickleSerializer
from datetime import datetime
import pytest

def test_string_serializer_serialize():
    """Test serialization of a string."""
    serializer = StringSerializer()
    data = "hello"
    serialized_data = serializer.serialize(data)
    assert serialized_data == b"hello"

def test_string_serializer_deserialize():
    """Test deserialization of bytes to string."""
    serializer = StringSerializer()
    data = b"hello"
    deserialized_data = serializer.deserialize(data)
    assert deserialized_data == "hello"

def test_string_serializer_round_trip():
    """Test round-trip serialization and deserialization."""
    serializer = StringSerializer()
    data = "hello"
    serialized_data = serializer.serialize(data)
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data
   

def test_json_serializer_serialize():
    """Test serialization of a dictionary."""
    serializer = JsonSerializer[dict]()
    data = {"key": "value"}
    serialized_data = serializer.serialize(data)
    assert serialized_data == b'{\n  "key": "value"\n}'

def test_json_serializer_deserialize():
    """Test deserialization of bytes to dictionary."""
    serializer = JsonSerializer[dict]()
    data = b'{"key": "value"}'
    deserialized_data = serializer.deserialize(data)
    assert deserialized_data == {'key': 'value'}

def test_json_serializer_round_trip():
    """Test round-trip serialization and deserialization."""
    serializer = JsonSerializer[dict]()
    data = {"key": "value"}
    serialized_data = serializer.serialize(data)
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data

def test_json_serializer_nested_dict():
    """Test serialization of a nested dictionary."""
    serializer = JsonSerializer[dict]()
    data = {"key": {"nested_key": "nested_value"}}
    serialized_data = serializer.serialize(data)
    assert serialized_data == b'{\n  "key": {\n    "nested_key": "nested_value"\n  }\n}'
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data

# jsonpickle

def test_jsonpickle_serializer_serialize():
    """Test serialization of a dictionary."""
    serializer = JsonPickleSerializer[dict]()
    data = {"key": "value"}
    serialized_data = serializer.serialize(data)
    assert serialized_data == b'{\n  "key": "value"\n}'

def test_jsonpickle_serializer_deserialize():
    """Test deserialization of bytes to dictionary."""
    serializer = JsonPickleSerializer[dict]()
    data = b'{"key": "value"}'
    deserialized_data = serializer.deserialize(data)
    assert deserialized_data == {"key": "value"}

def test_jsonpickle_serializer_round_trip():
    """Test round-trip serialization and deserialization."""
    serializer = JsonPickleSerializer[dict]()
    data = {"key": "value"}
    serialized_data = serializer.serialize(data)
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data

class CustomObject:
    def __init__(self, name: str, value: int) -> None:
        """Initialize CustomObject with name and value."""
        self.name: str = name
        self.value: int = value

    def get_string(self) -> str:
        """Return a string representation of the object."""
        var = 234
        return f"{self.name} - {self.value} - {var}"


def test_jsonpickle_serializer_nested_dict():
    """Test serialization of a nested dictionary."""
    serializer = JsonPickleSerializer[dict]()
    data = {"key": {"nested_key": "nested_value"}}
    serialized_data = serializer.serialize(data)
    assert serialized_data == b'{\n  "key": {\n    "nested_key": "nested_value"\n  }\n}'
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data


def test_jsonpickle_serializer_datetime():
    """Test serialization of a dictionary with datetime object."""
    serializer = JsonPickleSerializer[dict]()
    now = datetime.now()
    data = {"timestamp": now}
    serialized_data = serializer.serialize(data)
    # assert serialized_data == f'{{"timestamp": "{str(now)}"}}'.encode()
    deserialized_data = serializer.deserialize(serialized_data)
    assert deserialized_data == data


def test_jsonpickle_serializer_custom_object():
    """Test serialization of a custom object."""
    serializer = JsonPickleSerializer[dict]()
    obj = CustomObject(name="test", value=123)
    data = {"obj": obj}
    serialized_data = serializer.serialize(data)
    # assert serialized_data == b'{"name": "test", "value": 123}'
    deserialized_data = serializer.deserialize(serialized_data)

    assert isinstance(deserialized_data["obj"], CustomObject)
    assert deserialized_data["obj"].name == "test"
    assert deserialized_data["obj"].value == 123
    assert deserialized_data["obj"].get_string() == "test - 123 - 234"


import pytest
from pytest import Pytester

def test_snapshot_string(pytester: Pytester):
    """Test snapshot with string data."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_string(expect: Expect):
        expect.string("Hello World").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_fails_snapshot_not_exists(pytester: Pytester):
    """Test the failure of a snapshot when the snapshot data does not exist."""
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_string(expect: Expect):
        expect.string("Hello World").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v')
    assert result.ret == 1, "\n".join(result.outlines)

def test_fails_snapshot_string_mismatch(pytester: Pytester):
    """Test the failure of a snapshot when the snapshot data does not match."""
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_string(expect: Expect):
        expect.string("Hello World", name="string_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update', '--cache-clear')
    assert result.ret == 0, "\n".join(result.outlines)
    
    # Modify the test so it return a different value
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_string(expect: Expect):
        expect.string("Hello World!", name="string_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)

    result = pytester.runpytest('-v')
    assert result.ret == 1, "\n".join(result.outlines)
    result.stdout.fnmatch_lines([
        '*- Hello World!',
        '*+ Hello World',
    ])


def test_fails_snapshot_list_mismatch(pytester: Pytester):
    """Test the failure of a snapshot when the snapshot data does not match."""
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_string(expect: Expect):
        expect.list(["John Doe", 31]).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update', '--cache-clear')
    assert result.ret == 0, "\n".join(result.outlines)
    
    # Modify the test so it return a different value
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_string(expect: Expect):
        expect.list(["John Doe 2", 31]).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)

    result = pytester.runpytest("-v")
    assert result.ret == 1, "\n".join(result.outlines)
    assert 'E                 -   "John Doe 2",' in result.stdout.lines
    assert 'E                 +   "John Doe",' in result.stdout.lines
    # result.stdout.fnmatch_lines([
    #     '*- ["John Doe 2", 31]',
    #     '*+ ["John Doe", 31]',
    # ])

@pytest.mark.skip(reason="This fails under certain conditions. The second run of the tests do not get the new dict, I suspect it is due to a problem in pytester, not the snappylapy code. It is only a problem when the first element does not differ")
def test_fails_snapshot_dict2_mismatch(pytester: Pytester) -> None:
    """Test the failure of a snapshot when the dict snapshot data does not match."""
    test_code: str = """
    from snappylapy import Expect

    def test_fails_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update', '--cache-clear')
    assert result.ret == 0, "\n".join(result.outlines)

    # Delete the test_code.py file
    if not pytester.path.joinpath("test_code.py").exists():
        raise FileNotFoundError("test_code.py file not found.")
    (pytester.path / "test_code.py").unlink()

    # Modify the test so it returns a different value
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 32,
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)

    result = pytester.runpytest('-v', '-s', '--cache-clear')
    result.stdout.fnmatch_lines([
        '*- {"name": "John Doe 2", "age": 31}',
        '*+ {"name": "John Doe", "age": 31}',
    ])
    assert result.ret == 1, "\n".join(result.outlines)

def test_fails_snapshot_dict_mismatch(pytester: Pytester) -> None:
    """Test the failure of a snapshot when the dict snapshot data does not match."""
    test_code: str = """
    from snappylapy import Expect

    def test_fails_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update', '--cache-clear')
    assert result.ret == 0, "\n".join(result.outlines)

    # Delete the test_code.py file
    if not pytester.path.joinpath("test_code.py").exists():
        raise FileNotFoundError("test_code.py file not found.")
    (pytester.path / "test_code.py").unlink()

    # Modify the test so it returns a different value
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe 2",
            "age": 31,
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)

    result = pytester.runpytest('-v', '-s', '--cache-clear')
    result.stdout.fnmatch_lines([
        '*-   "name": "John Doe 2",',
        '*+   "name": "John Doe",',
    ])
    assert result.ret == 1, "\n".join(result.outlines)


def test_fails_snapshot_bytes_mismatch(pytester: Pytester):
    """Test the failure of a snapshot when the snapshot data does not match."""
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_bytes(expect: Expect):
        expect.bytes(b"Hello World", name="bytes_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)
    result = pytester.runpytest('-v', '--snapshot-update', '--cache-clear')
    assert result.ret == 0, "\n".join(result.outlines)
    
    # Modify the test so it return a different value
    test_code = """
    from snappylapy import Expect

    def test_fails_snapshot_bytes(expect: Expect):
        expect.bytes(b"Hello World!", name="bytes_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code=test_code)

    result = pytester.runpytest('-v')
    assert result.ret == 1, "\n".join(result.outlines)
    result.stdout.fnmatch_lines([
        '*- Hello World!',
        '*+ Hello World',
    ])
    print(result.outlines)

def test_snapshot_bytes(pytester: Pytester):
    """Test snapshot with bytes data."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_bytes(expect: Expect):
        expect.bytes(b"Hello World", name="bytes_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_dict(pytester: Pytester):
    """Test snapshot with dictionary data."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_list(pytester: Pytester):
    """Test snapshot with list data."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_list(expect: Expect):
        expect.list(["John Doe", 31]).to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_json_bytes(pytester: Pytester):
    """Test snapshot with JSON bytes data."""
    test_code = """
    import json
    from snappylapy import Expect

    def test_snapshot_json_bytes(expect: Expect):
        data = json.dumps({"name": "John Doe", "age": 31}).encode()
        expect.bytes(data, name="json_bytes_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_python_code(pytester: Pytester):
    """Test snapshot with Python code string."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_python_code(expect: Expect):
        py_code = "print('Hello World')"
        expect.string(py_code, filetype="py", name="python_code_snapshot").to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_with_custom_directories(pytester: Pytester):
    """Test snapshot with custom directories."""
    test_code = """
    import pathlib
    from snappylapy import Expect

    def test_snapshot_with_custom_directories(expect: Expect):
        expect.snapshot_dir = pathlib.Path("__snapshots_other_location__")
        expect.test_results_dir = pathlib.Path("__test_results_other_location__")
        expect.string("Hello World").to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)

def test_snapshot_multiple_assertions(pytester: Pytester):
    """Test snapshot with multiple assertions."""
    test_code = """
    from snappylapy import Expect

    def test_snapshot_multiple_assertions(expect: Expect):
        expect.string("Hello World").to_match_snapshot()
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)
    result = pytester.runpytest('-v')
    assert result.ret == 0, "\n".join(result.outlines)


def test_load_snapshot_from_file(pytester: Pytester):
    """Test loading snapshot data created in test_snapshot_dict from a file using the deserializer."""
    test_code = """
    from snappylapy import LoadSnapshot, Expect
    import pytest

    def test_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()

    @pytest.mark.snappylapy(depends=[test_snapshot_dict])
    def test_load_snapshot_from_file(load_snapshot: LoadSnapshot):
        data = load_snapshot.dict()
        assert data == {"name": "John Doe", "age": 31}
    """
    pytester.makepyfile(test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)


def test_load_snapshot_from_file_test_in_seperate_module(pytester: Pytester):
    """Check that the reordering of tests based on dependencies works."""
    create_snapshot_test_code = """
    from snappylapy import Expect

    def test_snapshot_dict(expect: Expect):
        expect.dict({
            "name": "John Doe",
            "age": 31
        }).to_match_snapshot()
    """
    load_snapshot_test_code = """
    from snappylapy import LoadSnapshot
    import test_create_snapshot
    import pytest

    @pytest.mark.snappylapy(depends=[test_create_snapshot.test_snapshot_dict])
    def test_load_snapshot_from_file(load_snapshot: LoadSnapshot):
        data = load_snapshot.dict()
        assert data == {"name": "John Doe", "age": 31}
    """

    pytester.makepyfile(test_create_snapshot=create_snapshot_test_code)
    pytester.makepyfile(test_a_load_snapshot=load_snapshot_test_code)
    result = pytester.runpytest('-v', '--snapshot-update')
    assert result.ret == 0, "\n".join(result.outlines)



"""Session for snapshot testing."""
import pathlib
from _pytest.terminal import TerminalReporter
from dataclasses import dataclass
from snappylapy.constants import DIRECTORY_NAMES


@dataclass
class SnapshotSession:
    """Session for snapshot testing."""

    def __init__(self) -> None:
        """Initialize the snapshot session."""
        self.snapshots_created: list[str] = []
        self.snapshots_updated: list[str] = []
        self.snapshot_tests_succeeded: list[str] = []
        self.snapshot_tests_failed: list[str] = []

    def _get_all_snapshots(self) -> set[str]:
        """Loop through all SNAPSHOT_DIR_NAME directories and return names of all snapshots."""
        snapshot_file_names: set[str] = set()
        # Find all directories called SNAPSHOT_DIR_NAME
        snapshot_dirs = list(pathlib.Path().rglob(
            DIRECTORY_NAMES.snapshot_dir_name))
        for snapshot_dir in snapshot_dirs:
            snapshot_file_names.update(
                snapshot_file.name for snapshot_file in snapshot_dir.iterdir())
        return snapshot_file_names

    def _get_unvisited_snapshots(self) -> set[str]:
        """Get all missing snapshots."""
        all_snapshots = self._get_all_snapshots()
        unvisited_snapshots = all_snapshots - set(
            self.snapshots_created + self.snapshots_updated +
            self.snapshot_tests_succeeded + self.snapshot_tests_failed)
        return unvisited_snapshots

    def on_finish(self) -> None:
        """On finish of the snapshot testing."""

    def has_ran_snapshot_tests(self) -> bool:
        """Check if snapshot tests have been run."""
        return bool(
            self.snapshots_created or self.snapshots_updated
            or self.snapshot_tests_succeeded)

    def write_summary(self, reporter: TerminalReporter) -> None:
        """Write the snapshot tests summary."""
        if not self.has_ran_snapshot_tests():
            return
        reporter.write_sep("=", "Snapshot tests summary", blue=True)
        if self.snapshot_tests_succeeded:
            reporter.write(
                f"Got {len(self.snapshot_tests_succeeded)} snapshot tests passing\n",
                green=True,
            )
        if self.snapshot_tests_failed:
            reporter.write(
                f"Got {len(self.snapshot_tests_failed)} snapshot tests failing\n",
                red=True,
            )
        if self.snapshots_updated:
            reporter.write("Updated snapshots:\n", green=True)
            for snapshot in self.snapshots_updated:
                reporter.write(f"  {snapshot}\n", blue=True)
        if self.snapshots_created:
            reporter.write("Created snapshots:\n", green=True)
            for snapshot in self.snapshots_created:
                reporter.write(f"  {snapshot}\n", blue=True)

        unvisited_snapshots = self._get_unvisited_snapshots()
        if unvisited_snapshots:
            reporter.write(
                f"Found {len(unvisited_snapshots)} unvisited snapshots:\n",
                red=True)
            for snapshot in unvisited_snapshots:
                reporter.write(f"  {snapshot}\n", blue=True)

    def add_created_snapshot(self, item: str) -> None:
        """Add a created snapshot."""
        self.snapshots_created.append(item)

    def add_updated_snapshot(self, item: str) -> None:
        """Add an updated snapshot."""
        self.snapshots_updated.append(item)

    def add_snapshot_test_succeeded(self, item: str) -> None:
        """Add a snapshot test that succeeded."""
        self.snapshot_tests_succeeded.append(item)

    def add_snapshot_test_failed(self, item: str) -> None:
        """Add a snapshot test that failed."""
        self.snapshot_tests_failed.append(item)



"""Exceptions for SnappyLapy."""


class TestDirectoryNotParametrizedError(Exception):
    """Custom exception raised when the test_directory fixture is not parametrized."""

    def __init__(self) -> None:
        """Initialize the exception with a custom error message."""
        error_msg = (
            "The test_directory fixture is not parametrized, please add the snappylapy marker to the test, "
            "e.g. @pytest.mark.snappylapy(foreach_folder_in='test_data')"
        )
        super().__init__(error_msg)



"""Create cli using the typer library."""

import re
import typer
import pathlib
import subprocess  # noqa: S404
from enum import Enum
from snappylapy._utils_directories import DirectoryNamesUtil
from snappylapy.constants import DIRECTORY_NAMES

app = typer.Typer(
    no_args_is_help=True,
    help="""
    Welcome to the snappylapy CLI!

    Use these commands to initialize your repository, update or clear test results and snapshots,
    and review differences between your test results and snapshots using the 'diff' command.

    - Run 'init' to set up your repo for snappylapy.
    - Use 'update' to refresh snapshots with the latest test results.
    - Use 'clear' to remove all test results and snapshots (add --force to skip confirmation).
    - Use 'diff' to view changes between test results and snapshots in your editor.

    For more details on each command, use --help after the command name.
    """,
)


@app.command()
def init() -> None:
    """
    Run this command to initialize your repository for snappylapy.

    This will add a line to your .gitignore file to ensure test results are not tracked by git.
    """
    # Check if .gitignore exists
    gitignore_path = pathlib.Path(".gitignore")
    if not gitignore_path.exists():
        typer.echo("No .gitignore file found. Creating one.")
        gitignore_path.touch()
    # Check if already in .gitignore
    with gitignore_path.open("r") as file:
        lines = file.readlines()
    regex = re.compile(rf"^{re.escape(DIRECTORY_NAMES.test_results_dir_name)}(/|$)")
    if any(regex.match(line) for line in lines):
        typer.echo("Already in .gitignore.")
        return
    # Add to .gitignore to top of file
    line_to_add = f"# Ignore test results from snappylapy\n{DIRECTORY_NAMES.test_results_dir_name}/\n\n"
    with gitignore_path.open("w") as file:
        file.write(line_to_add)
        file.writelines(lines)
    typer.echo(f"Added {DIRECTORY_NAMES.test_results_dir_name}/ to .gitignore.")


@app.command()
def clear(
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force deletion without confirmation",
    ),
) -> None:
    """
    Use this command to clear all test results and snapshots created by snappylapy.

    This will recursively delete all files and directories related to test results and snapshots.
    Use --force to skip confirmation.

    This finds and deletes all __test_results__ and __snapshots__ directories recursively across the working directory.
    """
    directories_to_delete = DirectoryNamesUtil().get_all_directories_created_by_snappylapy()
    list_of_files_to_delete = DirectoryNamesUtil().get_all_file_paths_created_by_snappylapy()
    if not list_of_files_to_delete:
        typer.echo("No files to delete.")
        return
    if not force:
        typer.echo("Deleting files:")
        for file in list_of_files_to_delete:
            typer.echo(f"- {file}")

        typer.secho(
            f"Deleting {len(list_of_files_to_delete)} files from {len(directories_to_delete)} directories:",
            fg=typer.colors.BRIGHT_BLUE,
        )
        for directory in directories_to_delete:
            typer.echo(f"- {directory}")

        # Ask for confirmation
        typer.secho("\nAre you sure you want to delete all test results and snapshots?", fg=typer.colors.BRIGHT_BLUE)
        response = typer.prompt("Type 'yes' to confirm, anything else to abort.", default="no")
        if response.lower() != "yes":
            typer.echo("Aborted.")
            return
    # Delete files
    delete_files(list_of_files_to_delete)
    typer.echo(f"Deleted {len(list_of_files_to_delete)} files.")


@app.command()
def update() -> None:
    """
    Use this command to update all snapshot files with the latest test results.

    This will overwrite existing snapshots with current test outputs, ensuring your snapshots reflect the latest changes.

    The file contents of any files in any of the __test_results__ folders will be copied to the corresponding __snapshots__ folder.
    """  # noqa: E501
    files_test_results = DirectoryNamesUtil().get_all_file_paths_test_results()
    if not files_test_results:
        typer.echo("No files to update.")
        return
    file_statuses = check_file_statuses(files_test_results)
    files_to_update = [file for file, status in file_statuses.items() if status != FileStatus.UNCHANGED]
    count_up_to_date_files = len(files_test_results) - len(files_to_update)
    if not files_to_update:
        typer.echo(f"All snapshot files are up to date. {count_up_to_date_files} files are up to date.")
        return

    typer.echo(
        f"Found {len(files_to_update)} files to update."
        + (f" {count_up_to_date_files} files are up to date." if count_up_to_date_files > 0 else ""),
    )
    for file in files_to_update:
        snapshot_file = file.parent.parent / DIRECTORY_NAMES.snapshot_dir_name / file.name
        snapshot_file.parent.mkdir(parents=True, exist_ok=True)
        snapshot_file.write_bytes(file.read_bytes())
        typer.echo(f"Updated snapshot: {snapshot_file}")


@app.command()
def diff() -> None:
    """
    Show the differences between the test results and the snapshots.

    Opens all of the changed diffs in the Visual Studio Code (VSCode) editor.
    This requires that you have VSCode installed and the `code` command available in your PATH.

    More diff viewers will be supported in the future, please raise a request on github with your needs.
    """
    files_test_results = DirectoryNamesUtil().get_all_file_paths_test_results()
    file_statuses = check_file_statuses(files_test_results)
    files_to_diff = [file for file, status in file_statuses.items() if status == FileStatus.CHANGED]
    if not files_to_diff:
        status_counts: dict[FileStatus, int] = dict.fromkeys(FileStatus, 0)
        for status in file_statuses.values():
            status_counts[status] += 1

        typer.secho("File status counts:", underline=True, bold=True)
        for status, count in status_counts.items():
            typer.echo(f"- {status.value}: {count} file(s)")
        typer.echo("No files have changed, not opening any diffs.")
        return
    typer.echo(f"Opening diffs for {len(files_to_diff)} changed files.")
    for file in files_to_diff:
        snapshot_file = file.parent.parent / DIRECTORY_NAMES.snapshot_dir_name / file.name
        success: bool = _try_open_diff(file, snapshot_file)
        if not success:
            typer.secho(
                f"Could not open diff tool. Files to compare:\n"
                f"  Test result: {file.resolve()}\n"
                f"  Snapshot:    {snapshot_file.resolve()}",
                fg=typer.colors.YELLOW,
            )


def delete_files(list_of_files_to_delete: list[pathlib.Path]) -> None:
    """Delete files."""
    # Delete files
    for file in list_of_files_to_delete:
        file.unlink()
    # Delete directories
    for dir_name in [
        DIRECTORY_NAMES.test_results_dir_name,
        DIRECTORY_NAMES.snapshot_dir_name,
    ]:
        for root_dir in pathlib.Path().rglob(dir_name):
            root_dir.rmdir()


def _try_open_diff(file1: pathlib.Path, file2: pathlib.Path) -> bool:
    """Try to open diff using available tools, return True if successful."""
    diff_commands: list[list[str]] = [
        ["code", "--diff", str(file1.resolve()), str(file2.resolve())],
        ["code.cmd", "--diff", str(file1.resolve()), str(file2.resolve())],  # Windows alternative
    ]

    for command in diff_commands:
        try:
            subprocess.run(command, check=True, timeout=10)  # noqa: S603 - shell=False and args as list, safe usage
        except subprocess.TimeoutExpired:  # noqa: PERF203
            typer.secho(
                f"Diff tool timed out for command: {' '.join(command)}",
                fg=typer.colors.RED,
            )
            continue
        except (subprocess.CalledProcessError, FileNotFoundError):
            continue
        else:
            return True
    return False


class FileStatus(Enum):
    """Enum to represent the status of a file."""

    NOT_FOUND = "not_found"
    CHANGED = "changed"
    UNCHANGED = "unchanged"


def check_file_statuses(
    file_paths: list[pathlib.Path],
) -> dict[pathlib.Path, FileStatus]:
    """Check the status of files in the snapshot directory."""
    file_statuses: dict[pathlib.Path, FileStatus] = {}
    for file_path in file_paths:
        snapshot_file = file_path.parent.parent / DIRECTORY_NAMES.snapshot_dir_name / file_path.name
        if not snapshot_file.exists():
            file_statuses[file_path] = FileStatus.NOT_FOUND
        elif snapshot_file.stat().st_size != file_path.stat().st_size:
            # TODO: This is not foolproof, does not catch content swaps and byte flips.
            file_statuses[file_path] = FileStatus.CHANGED
        elif snapshot_file.read_bytes() != file_path.read_bytes():
            # TODO: Expensive call, store hashes instead in a data file.
            file_statuses[file_path] = FileStatus.CHANGED
        else:
            file_statuses[file_path] = FileStatus.UNCHANGED
    return file_statuses


if __name__ == "__main__":
    app()



"""Make expectation classes available for import."""

from .expect_bytes import BytesExpect
from .expect_dataframe import DataframeExpect
from .expect_dict import DictExpect
from .expect_list import ListExpect
from .expect_object import ObjectExpect
from .expect_string import StringExpect

__all__ = [
    "BytesExpect",
    "DataframeExpect",
    "DictExpect",
    "ListExpect",
    "ObjectExpect",
    "StringExpect",
]



"""
Type resolution tests for `Expect.__call__` overloads.

These tests rely on `assert_type` so that static type checkers (mypy, pylance, pyright)
validate that each overload of `Expect.__call__` returns the correct *Expect class.
At runtime `assert_type` is a no-op returning the original value; we also add
`isinstance` assertions to guard dynamic dispatch.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import pytest

try:  # Python <3.11 fallback
    from typing import assert_type  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover
    from typing_extensions import assert_type  # type: ignore[attr-defined]

try:  # Optional dependency import for runtime test skipping logic
    import pandas as pd
except ImportError:  # pragma: no cover
    pd = None  # type: ignore[assignment]

from snappylapy.expectation_classes import (
    BytesExpect,
    DataframeExpect,
    DictExpect,
    ListExpect,
    StringExpect,
    ObjectExpect,
)
from snappylapy.fixtures import Expect
from snappylapy.models import Settings
from snappylapy.session import SnapshotSession


@pytest.fixture
def expect_fixture(tmp_path: Path) -> Expect:
    """Provide an `Expect` instance with minimal valid settings."""
    settings = Settings(
        test_filename="test_expect_type_overloads",
        test_function="test_func",
        snapshots_base_dir=tmp_path,
    )
    return Expect(SnapshotSession(), settings)


def test_overload_dict(expect_fixture: Expect) -> None:
    """Dict input returns DictExpect."""
    result = expect_fixture({"a": 1})
    assert isinstance(result, DictExpect)
    assert_type(result, DictExpect)


def test_overload_list(expect_fixture: Expect) -> None:
    """List input returns ListExpect."""
    result = expect_fixture([1, 2, 3])
    assert isinstance(result, ListExpect)
    assert_type(result, ListExpect)


def test_overload_string(expect_fixture: Expect) -> None:
    """String input returns StringExpect."""
    result = expect_fixture("hello")
    assert isinstance(result, StringExpect)
    assert_type(result, StringExpect)


def test_overload_bytes(expect_fixture: Expect) -> None:
    """Bytes input returns BytesExpect."""
    result = expect_fixture(b"data")
    assert isinstance(result, BytesExpect)
    assert_type(result, BytesExpect)


@pytest.mark.skipif(pd is None and not TYPE_CHECKING, reason="pandas not installed")
def test_overload_dataframe(expect_fixture: Expect) -> None:  # pragma: no cover - optional dependency
    """DataFrame input returns DataframeExpect (when pandas available)."""
    assert pd is not None  # for type checkers
    data_frame = pd.DataFrame({"a": [1, 2, 3]})
    result = expect_fixture(data_frame)
    assert isinstance(result, DataframeExpect)
    assert_type(result, DataframeExpect)

def test_overload_object(expect_fixture: Expect) -> None:
    """Object input returns ObjectExpect."""
    class Custom:
        pass

    result = expect_fixture(Custom())
    assert isinstance(result, ObjectExpect)
    assert_type(result, ObjectExpect)

@pytest.mark.parametrize(
    "data,expected_attr",
    [
        ({"a": 1}, "dict"),
        ([1, 2, 3], "list"),
        ("hello", "string"),
        (b"bytes", "bytes"),
    ],
)
def test_expect_selects_correct_class_for_builtin_types(data, expected_attr, expect_fixture):
    """Test that Expect.__call__ selects the correct class for builtin types."""
    # Patch all expectation attributes with mocks
    expect_fixture.dict = MagicMock(return_value="called_dict")
    expect_fixture.list = MagicMock(return_value="called_list")
    expect_fixture.string = MagicMock(return_value="called_string")
    expect_fixture.bytes = MagicMock(return_value="called_bytes")
    expect_fixture.dataframe = MagicMock(return_value="called_dataframe")
    expect_fixture.object = MagicMock(return_value="called_object")
    result = expect_fixture(data)
    assert result == f"called_{expected_attr}"
    getattr(expect_fixture, expected_attr).assert_called_once_with(data)


def test_expect_selects_object_for_custom_type(expect_fixture: Expect):
    """Test that Expect.__call__ falls back to object for custom types."""
    class Custom:
        pass

    custom_obj = Custom()
    expect_fixture.dict = MagicMock(return_value="called_dict")
    expect_fixture.list = MagicMock(return_value="called_list")
    expect_fixture.string = MagicMock(return_value="called_string")
    expect_fixture.bytes = MagicMock(return_value="called_bytes")
    expect_fixture.dataframe = MagicMock(return_value="called_dataframe")
    expect_fixture.object = MagicMock(return_value="called_object")
    result = expect_fixture(custom_obj)
    assert result == "called_object"
    expect_fixture.object.assert_called_once_with(custom_obj)



"""Base class for snapshot testing."""

from __future__ import annotations

import pathlib
import Levenshtein
from abc import ABC, abstractmethod
from snappylapy.models import Settings
from snappylapy.serialization import Serializer
from snappylapy.session import SnapshotSession
from typing import Generic, TypeVar

T = TypeVar("T")

LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE = 100
LARGE_DIFF_CHARACTER_THRESHOLD = 10_000


def calculate_difference_two_strings(a: str, b: str) -> int:
    """
    Approximate string difference.

    This is a gross approximation and may not reflect true semantic differences.
    """
    return Levenshtein.distance(a, b, score_cutoff=LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE)


class BaseSnapshot(ABC, Generic[T]):
    """Base class for snapshot testing."""

    serializer_class: type[Serializer[T]]

    def __init__(
        self,
        settings: Settings,
        snappylapy_session: SnapshotSession,
    ) -> None:
        """Initialize the base snapshot."""
        self.settings = settings
        self._data: T | None = None
        self.snappylapy_session = snappylapy_session

    @abstractmethod
    def __call__(
        self,
        data_to_snapshot: T,
        name: str | None = None,
        filetype: str = "snapshot.txt",
    ) -> BaseSnapshot[T]:
        """Prepare data for snapshot testing."""

    def to_match_snapshot(self) -> None:
        """Assert test results match the snapshot."""
        if not (self.settings.snapshot_dir / self.settings.filename).exists():
            if not self.settings.snapshot_update:
                error_msg = f"Snapshot file not found: {self.settings.filename}, run 'snappylapy update' command in the terminal, or run pytest with the --snapshot-update flag to create it."  # noqa: E501
                raise FileNotFoundError(error_msg)
            self.snappylapy_session.add_created_snapshot(self.settings.filename)
            self._update_snapshot()
            return

        snapshot_data = self._read_file(self.settings.snapshot_dir / self.settings.filename)
        test_data = self._read_file(self.settings.test_results_dir / self.settings.filename)
        try:
            self.compare_snapshot_data(snapshot_data, test_data)
        except AssertionError as error:
            if self.settings.snapshot_update:
                self.snappylapy_session.add_updated_snapshot(self.settings.filename)
                self._update_snapshot()
            else:
                self.snappylapy_session.add_snapshot_test_failed(self.settings.filename)
                diff_msg = str(error)
                error_msg = f"Snapshot does not match test results. Run pytest with the --snapshot-update flag to update the snapshot.\n{diff_msg}"  # noqa: E501
                raise AssertionError(error_msg)  # noqa: B904
        else:
            self.snappylapy_session.add_snapshot_test_succeeded(self.settings.filename)

    def compare_snapshot_data(self, snapshot_data: bytes, test_data: bytes) -> None:
        """
        Compare snapshot data with test data.

        Optimizations is needed for large data since difflib used in the pytest_assertion plugin rewrites are very slow
        when there are many diffs in large strings.
        """
        snapshot_data_str = snapshot_data.decode()
        snapshot_data_str = snapshot_data_str.replace("\r\n", "\n").replace("\r", "\n")
        test_data_str = test_data.decode()
        test_data_str = test_data_str.replace("\r\n", "\n").replace("\r", "\n")
        if (
            len(snapshot_data_str) > LARGE_DIFF_CHARACTER_THRESHOLD
            or len(test_data_str) > LARGE_DIFF_CHARACTER_THRESHOLD
        ):
            distance = Levenshtein.distance(
                snapshot_data_str,
                test_data_str,
                score_cutoff=LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE,
            )
            if distance < LARGE_DIFF_DISABLE_ASSERTION_REWRITES_ON_DISTANCE:
                assert snapshot_data_str == test_data_str
            if snapshot_data_str != test_data_str:
                msg = (
                    "Snapshots are not matching. Data is too large to show differences. "
                    "Instead use the 'snappylapy diff' command to see the differences."
                )
                raise AssertionError(msg)
        else:
            assert snapshot_data_str == test_data_str

    def _prepare_test(self, data: T, name: str | None, extension: str) -> None:
        """Prepare and save test results."""
        if name is not None:
            self.settings.custom_name = str(name)
        self._data = data
        self.settings.filename_extension = extension
        file_path = self.settings.test_results_dir / self.settings.filename
        file_path.parent.mkdir(parents=True, exist_ok=True)
        self._save_test_results(file_path, data)

    def _update_snapshot(self) -> None:
        """Write test results to the snapshot file."""
        snap_path = self.settings.snapshot_dir / self.settings.filename
        test_path = self.settings.test_results_dir / self.settings.filename
        snap_path.parent.mkdir(parents=True, exist_ok=True)
        snap_path.write_bytes(test_path.read_bytes())

    def _read_file(self, path: pathlib.Path) -> bytes:
        """Read file bytes or return placeholder."""
        return path.read_bytes() if path.exists() else b"<No file>"

    def _save_test_results(self, path: pathlib.Path, data: T) -> None:
        """Save data for test results."""
        data_bin = self.serializer_class().serialize(data)
        path.write_bytes(data_bin)



"""Snapshot testing and expectations for generic custom objects."""
from __future__ import annotations

from .base_snapshot import BaseSnapshot
from snappylapy.serialization import JsonPickleSerializer


class ObjectExpect(BaseSnapshot[object]):
    """Snapshot testing for generic objects."""

    serializer_class = JsonPickleSerializer[object]

    def __call__(
        self,
        data_to_snapshot: object,
        name: str | None = None,
        filetype: str = "object.json",
    ) -> ObjectExpect:
        """Prepare an object for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self



"""
Serialization classes for serializing and deserializing data.

Be sure that data is serialized the same way no matter what os and OS configuration is used.
"""
import json
import jsonpickle
from abc import ABC, abstractmethod
from snappylapy.constants import OUTPUT_JSON_INDENTATION_LEVEL
from typing import Generic, TypeVar

T = TypeVar("T")

ENCODING_TO_USE = "utf-8"


class Serializer(ABC, Generic[T]):
    """Base class for serialization."""

    @abstractmethod
    def serialize(self, data: T) -> bytes:
        """Serialize data to bytes."""

    @abstractmethod
    def deserialize(self, data: bytes) -> T:
        """Deserialize bytes to data."""


class JsonSerializer(Serializer, Generic[T]):
    """Serialize and deserialize a dictionary."""

    def serialize(self, data: T) -> bytes:
        """Serialize a dictionary to bytes with cross-platform consistency."""
        json_string = json.dumps(
            data,
            default=str,
            indent=OUTPUT_JSON_INDENTATION_LEVEL,
            ensure_ascii=False,
        )
        json_string = json_string.replace("\r\n", "\n").replace("\r", "\n")  # Normalize all line endings to LF
        return json_string.encode(encoding=ENCODING_TO_USE)

    def deserialize(self, data: bytes) -> T:
        """Deserialize bytes to a dictionary."""
        return json.loads(data.decode(encoding=ENCODING_TO_USE))


class JsonPickleSerializer(Serializer, Generic[T]):
    """Serialize and deserialize a dictionary using pickle."""

    def serialize(self, data: T) -> bytes:
        """Serialize a dictionary/list or other to bytes in json format with cross-platform consistency."""
        json_string: str = jsonpickle.encode(
            data,
            indent=OUTPUT_JSON_INDENTATION_LEVEL,
        )
        json_string = json_string.replace("\r\n", "\n").replace("\r", "\n")  # Normalize all line endings to LF
        return json_string.encode(encoding=ENCODING_TO_USE)

    def deserialize(self, data: bytes) -> T:
        """Deserialize bytes to a dictionary or list."""
        return jsonpickle.decode(data.decode(encoding=ENCODING_TO_USE))  # noqa: S301, pickle security, data should be trusted here, keep your snapshot files safe


class StringSerializer(Serializer[str]):
    """Serialize and deserialize a string."""

    def serialize(self, data: str) -> bytes:
        """Serialize a string to bytes."""
        return data.encode(encoding=ENCODING_TO_USE)

    def deserialize(self, data: bytes) -> str:
        """Deserialize bytes to a string."""
        return data.decode(encoding=ENCODING_TO_USE)


class BytesSerializer(Serializer[bytes]):
    """Serialize and deserialize bytes."""

    def serialize(self, data: bytes) -> bytes:
        """Already in bytes, return as is."""
        return data

    def deserialize(self, data: bytes) -> bytes:
        """Already in bytes, return as is."""
        return data



# ruff: noqa: I001, E402
"""Define the public api for the snappylapy package."""
import pytest
pytest.register_assert_rewrite("snappylapy.expectation_classes.base_snapshot")
from .fixtures import Expect, LoadSnapshot


__all__ = ["Expect", "LoadSnapshot"]



"""Snapshot testing and expectations for dataframes."""

from __future__ import annotations

import sys
from .base_snapshot import BaseSnapshot
from functools import wraps
from snappylapy.serialization import JsonPickleSerializer
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if TYPE_CHECKING:
    import pandas as pd

F = TypeVar("F", bound=Callable[..., Any])


def require_pandas(func: F) -> F:
    """Decorate to require pandas for the function."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        try:
            import pandas as pd  # noqa: F401, PLC0415
        except ImportError as exc:
            error_message = "pandas is required for this function."
            raise ImportError(error_message) from exc
        return func(*args, **kwargs)

    return cast("F", wrapper)


class DataframeExpect(BaseSnapshot["pd.DataFrame"]):
    """Snapshot testing for dataframes."""

    serializer_class = JsonPickleSerializer["pd.DataFrame"]
    DataFrame: TypeAlias = "pd.DataFrame"

    @require_pandas
    def __call__(
        self,
        data_to_snapshot: "pd.DataFrame",  # noqa: UP037
        name: str | None = None,
        filetype: str = "dataframe.json",
    ) -> DataframeExpect:
        """Prepare a dataframe for snapshot testing."""
        self._prepare_test(data_to_snapshot, name, filetype)
        return self

    @require_pandas
    def column_not_to_contain_nulls(
        self,
        column_name: str,
    ) -> DataframeExpect:
        """Check that a column does not contain null values."""
        if self._data is None:
            error_message = "No data to check. Call __call__ first."
            raise ValueError(error_message)
        if column_name not in self._data.columns:
            error_message = f"Column {column_name} not found in dataframe."
            raise ValueError(error_message)
        if self._data[column_name].isna().any():
            null_count: int = int(self._data[column_name].isna().sum())
            top_n: int = 3  # Number of examples to display
            null_rows: pd.DataFrame = self._data[self._data[column_name].isna()].head(top_n)
            null_value_error_message = (
                f"Column {column_name} contains {null_count} null values. "
                f"Top {top_n} examples:\n{null_rows.to_string(index=False)}"
            )
            raise ValueError(
                null_value_error_message,
            )
        return self

    @require_pandas
    def columns_not_to_contain_nulls(
        self,
        column_names: list[str] | None = None,
    ) -> DataframeExpect:
        """Check that multiple columns do not contain null values."""
        if self._data is None:
            error_message = "No data to check. Call __call__ first."
            raise ValueError(error_message)
        if not column_names:
            # Check all columns
            column_names = self._data.columns.tolist()
        assert column_names is not None, "Column names should have been set to default to all columns."
        error_texts: list[str] = []
        for column_name in column_names:
            try:
                self.column_not_to_contain_nulls(column_name)
            except ValueError as e:  # noqa: PERF203
                error_texts.append(str(e))
        if error_texts:
            raise ValueError("\n".join(error_texts))
        return self

    @require_pandas
    def columns_to_match_regex(
        self,
        column_to_regex: dict[str, str],
    ) -> DataframeExpect:
        """
        Check that columns match regex.

        column_to_regex: dict[str, str]
            Dictionary of column names to regex patterns.
            The column name is the key and the regex pattern is the value.
        """
        if self._data is None:
            error_message = "No data to check. Call __call__ first."
            raise ValueError(error_message)
        for column_name, regex in column_to_regex.items():
            if column_name not in self._data.columns:
                error_message = f"Column {column_name} not found in dataframe."
                raise ValueError(error_message)
            if not self._data[column_name].astype(str).str.match(regex).all():
                regex_mismatch_message = (
                    f"Column {column_name} does not match regex {regex}. "
                    f"Top 3 examples:\n{self._data[column_name].head(3).to_string(index=False)}"
                )
                raise ValueError(
                    regex_mismatch_message,
                )
        return self
